package com.biganto.visual.roompark.data.data_provider

import com.biganto.visual.androidplayer.data.repository.local.ILocalStore
import com.biganto.visual.roompark.data.local.UserState
import com.biganto.visual.roompark.data.repository.api.room_park.IRoomParkApi
import com.biganto.visual.roompark.data.repository.db.IDb
import com.biganto.visual.roompark.data.repository.db.requrey.model.DealEntity
import com.biganto.visual.roompark.data.repository.db.requrey.model.EstateEntity
import com.biganto.visual.roompark.data.repository.db.requrey.model.UserEntity
import com.biganto.visual.roompark.data.repository.mapper.fromRaw
import com.biganto.visual.roompark.data.repository.mapper.fromRawList
import com.biganto.visual.roompark.domain.contract.DealContract
import com.biganto.visual.roompark.domain.custom_exception.CustomApiException
import com.biganto.visual.roompark.domain.model.EstateModel
import com.biganto.visual.roompark.domain.model.fromEntity
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.rxkotlin.Observables
import io.reactivex.schedulers.Schedulers
import timber.log.Timber
import javax.inject.Inject

/**
 * Created by Vladislav Bogdashkin on 29.10.2019.
 */
//


class EstateRepository @Inject constructor(
    local: ILocalStore,
    private val api: IRoomParkApi,
    private val db: IDb
): DealContract {


    init {
        Timber.d("Estate Repository Created $this")
    }

    override fun fetchFavorites(user:UserEntity): Observable<List<EstateEntity>> =
        Observables.zip(
        api.getFavorites(user.authToken)
            .doOnError(Timber::e)
            .map { fromRawList(it, ::fromRaw) }
            ,db.getUserFavorites(user.uuid)
                .toList().toObservable()
                .doOnError(Timber::e)
        ){apiList,dbList ->
            apiList.forEach { estate ->
                estate.setFavorite(true)
                estate.user = user
            }
            dbList
                .filterNotNull()
                .filter { dbFav ->!apiList.map{it.id}.contains(dbFav.id) }
                .let{ db.deleteEstate(it) }

            return@zip apiList
        }
            .flatMap(db::upsertEstates)
            .map { it.toList() }
            .doOnNext{ db.refreshUser(user) }

    override fun fetchDeals(user:UserEntity): Observable<List<DealEntity>> =
        getDealsApi(user)


    private fun fetchEstateDb(id: Int) = db.getEstate(id)

    override fun getEstate(estateId: Int): Observable<EstateModel> {
        return fetchEstateDb(estateId).map(::fromEntity)
    }



    override fun getDealsApi(user: UserEntity): Observable<List<DealEntity>> =
        Observables.zip(
            api.getDeals(user.authToken)
                .doOnError(Timber::e)
                .map {  List(it.size){index -> fromRaw(it[index],user)} }
            ,db.getUserDeals(user.uuid)
                .toList().toObservable()
                .doOnError(Timber::e)
        ){apiList,dbList ->
            apiList.forEach { deal ->
                dbList?.firstOrNull { dbDeal-> dbDeal?.id == deal.id }?.let {
                    deal.setRead(it.read)
                }
                deal.user = user
            }
            dbList
                .filterNotNull()
                .filter { dbFav ->!apiList.map{it.id}.contains(dbFav.id) }
                .let{ db.deleteDeal(it) }
            return@zip apiList
        }
            .doOnNext(db::blockingUpsert)
            .doOnNext { db.refreshUser(user) }

    private val getDealsDb: Observable<List<DealEntity>> =
        local.recentUser()
            .flatMap {
                when (it) {
                    is UserState.Authenticated -> db.fetchUser(it.uuid.toInt()).take(1)
                    else -> throw CustomApiException.NotAuthorizedException()
                }
            }
            .map { it.deals?.map {deal ->  deal as DealEntity } }




    override fun setDealRead(dealId: String): Completable =
        db.setDealReadState(dealId,true)
            .doOnError { Timber.e(it) }




    private fun fetchEstateApi(building: Int,number:Int)=
        api.getEstate(building,number)
            .doOnNext { Timber.d("raw0 $it") }
            .map (::fromRaw)
            .doOnNext(db::upsertEstate)
            .subscribeOn(Schedulers.io())

    private fun fetchEstateDb(building: Int,number:Int) =
        db.fetchEstateByNumber(building,number.toString())
            .observable()
            .subscribeOn(Schedulers.io())


    override fun fetchEstate(building: Int,number:Int): Observable<EstateModel> =
        Observable.mergeDelayError(
            arrayListOf(fetchEstateApi(building,number),
                fetchEstateDb(building,number)
            )
        ).map (::fromEntity)
}
